import dummy_local_file  # pylint:disable=import-error,unused-import
from dagster_tests.cli_tests.command_tests.test_cli_commands import (  # pylint:disable=unused-import
    foo_pipeline,
    qux_job,
)
